.field-list ul {
  padding-left: 2em;
}

.bd-article {
  padding-top: 0.75rem;
}

.bd-content h1 {
  margin-top: 0;
}

.bd-content h2 {
  border-bottom: 1px solid var(--pst-color-border);
  padding-bottom: 0.3rem;
}
/* Light and Dark Theme Specific Styles */
/*
html[data-theme="light"] .bd-content h2 {
  border-bottom: 1px solid #d0d0d0;
}

html[data-theme="dark"] .bd-content h2 {
  border-bottom: 1px solid #444;
}
*/

:root {
  --pst-font-size-h1: 2.0rem;
  --pst-font-size-h2: 1.6rem;
  --pst-font-size-h3: 1.35rem;
  --pst-font-size-h4: 1.15rem;
  --pst-font-size-h5: 1.0rem;
  --pst-font-size-h6: 0.9rem;
}

.section {
  margin-bottom: 1.6rem;
}

p {
  margin-bottom: 0.7rem;
}
