# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Subpackage containing tools for detecting point-like (stellar) sources
or local peaks in an astronomical image.
"""

from .core import *  # noqa: F401, F403
from .daofinder import *  # noqa: F401, F403
from .irafstarfinder import *  # noqa: F401, F403
from .peakfinder import *  # noqa: F401, F403
from .starfinder import *  # noqa: F401, F403
